from ether import abi, crypto

from typing import Any, Dict, List
from ether.ether_types import EthABI


def make_selector(f: Dict[str, Any]) -> bytes:
    '''
    Parses a function ABI into a function selector
    This is the first 4 bytes of the hash of the signature

    Args:
        f (dict): the function ABI
    Returns:
        (bytes): the function selector
    '''
    return crypto.keccak256(abi.make_signature(f).encode('utf8'))[:4]


def _encode_function_args(
        function: Dict[str, Any],
        function_args: List[Any]) -> bytes:
    '''
    encodes function arguments into a data blob
    This gets prepended with the function selector
    '''
    return abi.encode_tuple(
        abi._make_type_tuple(function),
        function_args)


def _find_by_name(
        function_name: str,
        num_args: int,
        abi: EthABI) -> List[Dict[str, Any]]:
    '''
    Find a function within the ABI based on
    a) its name
    b) the number of args
    c) whether the args provided match that function's interface
    '''
    funcs = [e for e in abi if e['type'] == 'function']
    funcs = [f for f in funcs if f['name'] == function_name]
    funcs = [f for f in funcs if len(f['inputs']) == num_args]

    if len(funcs) == 0:
        raise ValueError('no functions with acceptable interface')

    return funcs


def encode_call(function: Dict[str, Any], function_args: List[Any]) -> bytes:
    '''
    Makes the data blob for a solidity contract function call
    This is a 4 byte selector, and then any number of bytes
    '''
    return (make_selector(function)
            + _encode_function_args(function, function_args))


def call(
        function_name: str,
        function_args: List[Any],
        contract_abi: EthABI) -> bytes:
    '''
    Call a function by name
    '''
    functions = _find_by_name(function_name, len(function_args), contract_abi)
    for function in functions:
        # return the first one that works
        try:
            return encode_call(function, function_args)
        except abi.ABIEncodingError:
            continue
    raise ValueError('no functions with acceptable interface')



from ether import abi, crypto

from typing import Any, cast, Dict, List
from ether.ether_types import EthABI, ParsedEtherEvent, UnparsedEtherEvent


def _make_topic0(event: Dict[str, Any]) -> str:
    '''
    Calculates the event topic hash frrom the event ABI
    Args:
        event (dict): the event ABI
    Returns:
        (str): the event topic as 0x prepended hex
    '''
    signature = abi.make_signature(event).encode('utf8')
    topic_hex = crypto.keccak256(signature).hex()
    return '0x{}'.format(topic_hex)


def _match_topic0_to_event(
        event_topic: str,
        events: List[Dict[str, Any]]) -> Dict[str, Any]:
    '''
    Finds the corresponding event from a topic string
    Args:
        event_topic (str): the event's 0x prepended hex topic
    Returns:
        (dict): the event ABI
    '''
    for event in events:
        if _make_topic0(event) == event_topic:
            return event
    raise ValueError('Topic not found')


def _find_indexed(event: Dict[str, Any]) -> List[Dict[str, Any]]:
    '''
    Finds indexed arguments
    Args:
        event_topic (str): the event's 0x prepended hex topic
    Returns:
        (list): the indexed arguments
    '''
    return [t for t in event['inputs'] if t['indexed']]


def _find_unindexed(event: Dict[str, Any]) -> List[Dict[str, Any]]:
    '''
    Finds indexed arguments
    Args:
        event_topic (str): the event's 0x prepended hex topic
    Returns:
        (list): the unindexed arguments
    '''
    return [t for t in event['inputs'] if not t['indexed']]


def decode_event(
        encoded_event: UnparsedEtherEvent,
        contract_abi: EthABI) -> Dict[str, Any]:
    '''
    Decodes an event using the provided abi
    Args:
        encoded_event (dict): the etherscan/full node event Dict
        abi           (dict): the abi as a dict (use json.loads)
    Returns:
        (dict): the decoded dict
    '''
    ret = {}

    events = [entry for entry in contract_abi if entry['type'] == 'event']

    # find the appropriate event interface
    event_abi = _match_topic0_to_event(encoded_event['topics'][0], events)

    # get the indexed args
    indexed = _find_indexed(event_abi)
    for signature, blob in zip(indexed, encoded_event['topics'][1:]):
        val = abi.decode(signature['type'], bytes.fromhex(blob[2:]))
        ret[signature['name']] = val

    unindexed = _find_unindexed(event_abi)
    unindexed_values = abi.decode_many(
        [t['type'] for t in unindexed],
        bytes.fromhex(encoded_event['data'][2:]))

    for k, v in zip([t['name'] for t in unindexed], unindexed_values):
        ret[k] = v

    ret['event_name'] = event_abi['name']

    return ret


def parse_event_data(
        encoded_event: UnparsedEtherEvent,
        contract_abi: EthABI) -> ParsedEtherEvent:
    '''Parses an event given a contract ABI'''
    tmp = cast(ParsedEtherEvent, encoded_event.copy())
    tmp['data'] = decode_event(encoded_event, contract_abi)
    return tmp



from ether import crypto, rlp

from ether.ether_types import EthSig, EthTx, SignedEthTx, UnsignedEthTx

from typing import cast, List


def hash_to_sign(tx: EthTx) -> bytes:
    '''Get the hash to sign. keccak256 of serialized tx with chainId'''
    return crypto.keccak256(serialize(strip_signature(tx)))


def strip_signature(tx: EthTx) -> UnsignedEthTx:
    '''Strip the signature from a tx. Recover chainId from v if necessary'''
    tmp_tx = cast(dict, tx).copy()
    if 'v' in tx:
        tmp_tx.pop('r')
        tmp_tx.pop('s')
        v = tmp_tx.pop('v')
        tmp_tx['chainId'] = (v - (35 if v % 2 else 36)) // 2
    return cast(UnsignedEthTx, tmp_tx)


def get_signature(tx: UnsignedEthTx, key: bytes) -> EthSig:
    '''Get a signature on a tx under a certain key'''
    if not tx['chainId']:
        raise ValueError('This library enforces EIP-155.')
    return crypto.sign(hash_to_sign(tx), key)


def sign_transaction(
        tx: UnsignedEthTx,
        key: bytes) -> SignedEthTx:
    '''
    Signs an ethereum transaction with a specified key
    '''
    sig = get_signature(tx, key)

    return SignedEthTx(
        to=tx['to'],
        value=tx['value'],
        gas=tx['gas'],
        gasPrice=tx['gasPrice'],
        nonce=tx['nonce'],
        data=tx['data'],
        v=sig[0], r=sig[1], s=sig[2])


def serialize_hex(t: EthTx) -> str:
    return f'0x{serialize(t).hex()}'


def serialize(t: EthTx) -> bytes:
    '''
    serialize a signed Ethereum transaction
    '''
    tx = cast(dict, t)
    v = tx['v'] if 'v' in tx else tx['chainId']
    r = tx['r'] if 'r' in tx else 0
    s = tx['s'] if 's' in tx else 0

    to_serialize: List[bytes] = [
        rlp.i2be_rlp_padded(tx['nonce']),
        rlp.i2be_rlp_padded(tx['gasPrice']),
        rlp.i2be_rlp_padded(tx['gas']),
        bytes.fromhex(tx['to'][2:]),
        rlp.i2be_rlp_padded(tx['value']),
        tx['data'],
        rlp.i2be_rlp_padded(v),
        rlp.i2be_rlp_padded(r),
        rlp.i2be_rlp_padded(s)
    ]

    return rlp.encode(to_serialize)


def deserialize(raw: bytes) -> EthTx:
    decoded = cast(List[bytes], rlp.decode(raw))

    # post-processing step
    nonce = rlp.be2i_rlp(decoded[0])
    gas_price = rlp.be2i_rlp(decoded[1])
    gas = rlp.be2i_rlp(decoded[2])
    to = f'0x{decoded[3].hex()}'
    value = rlp.be2i_rlp(decoded[4])
    data = decoded[5]
    (v, r, s) = [rlp.be2i_rlp(i) for i in decoded[6:]]

    tx = {
        'nonce': nonce,
        'gasPrice': gas_price,
        'gas': gas,
        'to': to,
        'value': value,
        'data': data
    }

    if r == 0:  # unsigned, but serialized with chainId as v
        tx['chainId'] = v
        return cast(UnsignedEthTx, tx)

    # signed
    tx['v'] = v
    tx['r'] = r
    tx['s'] = s
    return cast(SignedEthTx, tx)


def deserialize_hex(hex_raw: str) -> EthTx:
    h = hex_raw if hex_raw[:2] != '0x' else hex_raw[2:]
    return deserialize(bytes.fromhex(h))


def recover_sender(tx: SignedEthTx) -> str:
    pubkey = crypto.recover_pubkey(
        (tx['v'], tx['r'], tx['s']),
        hash_to_sign(tx))
    return crypto.pub_to_addr(pubkey)
